/*
                                  makeit.c

*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*
*       Part of:        Swarp
*
*       Author:         E.BERTIN (IAP)
*
*       Contents:       Main loop
*
*       Last modify:    21/11/2003
*
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#ifdef HAVE_CONFIG_H
#include	"config.h"
#endif

#ifdef HAVE_MATHIMF_H
#include <mathimf.h>
#else
#include <math.h>
#endif
#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_MPI
#include <mpi.h>
#endif

#include "define.h"
#include "globals.h"
#include "fits/fitscat.h"
#include "fitswcs.h"
#include "back.h"
#include "data.h"
#include "field.h"
#include "header.h"
#include "prefs.h"
#include "resample.h"

#ifdef HAVE_MPI
void	control_mpi(int nnodes, int ntasks);
#endif

#define	NFIELD	128	/* Increment in the number of fields */

/********************************** makeit ***********************************/
void	makeit(void)
  {
   fieldstruct		**infield,**inwfield, *outfield,*outwfield;
   catstruct		*cat, *wcat;
   tabstruct		*tab;
   double		w,w1,w2, invscale;
   static int	       	list[MAXINFIELD],next[MAXINFIELD];
   int			i,j,k, ninfield, ntinfield, lng,lat, forceflag, nfield;
   int			mpiflag, mpitask;

/* Install the signal-catching routines for temporary file cleanup */
#ifdef USE_THREADS
  install_cleanup(cancel_resample_threads);
#else
  install_cleanup(NULL);
#endif
/* Load input images */
  ninfield = prefs.ninfield;
/* First check input files and count FITS extensions when available */
  ntinfield = k = 0;
  nfield = NFIELD;
  QMALLOC(infield, fieldstruct *, nfield);
  QMALLOC(inwfield, fieldstruct *, nfield);
  NFPRINTF(OUTPUT, "Examining input data...")
  for (i=0; i<ninfield; i++)
    {
    if (!(cat=read_cat(prefs.infield_name[i])))
      {
      sprintf(gstr, "*Error*: %s not found", prefs.infield_name[i]);
      error(EXIT_FAILURE, gstr,"");
      }
/*-- Examine all extensions */
    wcat = NULL;
    if (prefs.weight_type[i] && prefs.weight_type[i] != WEIGHT_FROMBACK)
      if (!(wcat=read_cat(prefs.inwfield_name[i])))
        {
        sprintf(gstr, "*Error*: %s not found", prefs.inwfield_name[i]);
        error(EXIT_FAILURE, gstr,"");
        }
    tab=cat->tab;
    for (j=0; j<cat->ntab; j++)
      {
      if (tab->naxis
	&& (!(tab->tfields && tab->bitpix==8 && tab->pcount)))
        {
        if (k >= nfield)
	  {
          nfield += NFIELD;
          QREALLOC(infield, fieldstruct *, nfield);
          QREALLOC(inwfield,fieldstruct *, nfield);
          }
        list[k] = j;
        infield[k] = load_field(cat, list[k], i);
/*------ Set flags */
        inwfield[k] = wcat?
		load_weight(wcat, infield[k], list[k], i, prefs.weight_type[i])
		: NULL;
        next[i]++;
        k++;
        }
      tab = tab->nexttab;
      }
    ntinfield += next[i];
    if (!next[i])
      warning("No suitable data found in ", cat->filename);
    free_cat(&cat, 1);
    if (wcat)
      free_cat(&wcat, 1);
    }

/* Create output image (but nothing written to disk yet) */
  outwfield = NULL;
  NFPRINTF(OUTPUT, "Creating NEW output image...")
  outfield = init_field(infield, ntinfield, prefs.outfield_name);
/* Check consistency of lng and lat axes */
  lng = outfield->wcs->lng;
  lat = outfield->wcs->lat;
  forceflag = 0;
  if (lng != -1 && lat != -1)
    for (i=0; i<ntinfield; i++)
      if (infield[i]->wcs->lng != lng || infield[i]->wcs->lat != lat)
        {
/*------ Change axis labeling of discrepant wcs */
        reaxe_wcs(infield[i]->wcs, lng, lat);
        forceflag++;
        }
  if (forceflag)
    {
    end_field(outfield);
    outfield = init_field(infield, ntinfield, prefs.outfield_name);
    }

  NFPRINTF(OUTPUT, "")
  QPRINTF(OUTPUT, "------- Output File %s:\n", outfield->rfilename);
  printinfo_field(outfield, outwfield);
  NFPRINTF(OUTPUT, "Creating NEW weight-map...")
  outwfield = init_weight(prefs.outwfield_name, outfield);

/* HEADER_ONLY option: write the output FITS header and exit */
  if (prefs.headeronly_flag)
    {
/*-- Open output file and save header */
    if (open_cat(outfield->cat, WRITE_ONLY) != RETURN_OK)
      error(EXIT_FAILURE, "*Error*: cannot open for writing ",
		outfield->filename);
/*-- Add relevant information to FITS header */
    if (prefs.node_index==0)
      {
      writefitsinfo_outfield(outfield, *infield);
      QFWRITE(outfield->tab->headbuf, outfield->tab->headnblock*FBSIZE,
	outfield->cat->file, outfield->filename);
      }
    goto the_end;
    }

/* Compute projected limits and scaling in output frame (avoid unnecessary */
/* computations) */
  if (prefs.resample_flag || !prefs.combine_flag)
    {
    NFPRINTF(OUTPUT, "Framing and scaling images...")
    for (i=0; i<ntinfield; i++)
      {
      frame_wcs(infield[i]->wcs, outfield->wcs);
      if (prefs.fscalastro_type==FSCALASTRO_FIXED)
        scale_field(infield[i], outfield);
      }
    }

#ifdef HAVE_MPI
  if (!prefs.node_index)
    control_mpi(prefs.nnodes, ntinfield);
  else
#endif
    {
/*-- Read and transform the data */
    NFPRINTF(OUTPUT, "Loading input data...")
    k = 0;
    mpiflag = 1;
    mpitask = 0;
    for (i=0; i<ninfield; i++)
      {
      for (j=0; j<next[i]; j++, k++)
        {
#ifdef HAVE_MPI
        if (mpiflag)
          {
          MPI_Send(&mpitask, 1 , MPI_INT, 0, prefs.node_index, MPI_COMM_WORLD);
          MPI_Recv(&mpitask, 1 , MPI_INT, 0, prefs.node_index, MPI_COMM_WORLD,
		MPI_STATUS_IGNORE);
          mpiflag = 0;
          }
        if (mpitask<0 || k != mpitask)
          continue;
        mpiflag = 1;
#endif
/*------ Display some info */
        if (!j)
          {
          NFPRINTF(OUTPUT, "")
          QPRINTF(OUTPUT, "-------------- File %s:\n", infield[k]->rfilename);
          }
        printinfo_field(infield[k], inwfield[k]);

        if (prefs.resample_flag)
          {
/*-------- Open input files */
          if (open_cat(infield[k]->cat, READ_ONLY) != RETURN_OK)
            error(EXIT_FAILURE, "*Error*: Cannot re-open ",
		infield[k]->filename);
          if (inwfield[k])
            {
            if (open_cat(inwfield[k]->cat, READ_ONLY) != RETURN_OK)
              error(EXIT_FAILURE, "*Error*: Cannot re-open ",
		inwfield[k]->filename);
            }
/*-------- Pre-compute the background map */
          make_back(infield[k], inwfield[k]);
          }
        if (inwfield[k])
          sprintf(gstr, "   Weight scale: %.7g", inwfield[k]->sigfac);
        else
          *gstr = '\0';
        NPRINTF(OUTPUT, "    Background: %.7g   RMS: %.7g%s\n",
		infield[k]->backmean, infield[k]->backsig, gstr);

        if (prefs.resample_flag)
          {
/*-------- Read (and convert) the weight data */
          if (inwfield[k])
            {
            sprintf(gstr, "Reading %s", inwfield[k]->filename);
            NFPRINTF(OUTPUT, gstr)
            read_weight(inwfield[k]);
            }
/*-------- Read (and convert) the data */
          sprintf(gstr, "Reading %s", infield[k]->filename);
          NFPRINTF(OUTPUT, gstr)
          read_data(infield[k], inwfield[k]);
/*-------- Resample the data (no need to close catalogs) */
          sprintf(gstr, "Resampling %s", infield[k]->filename);
          NFPRINTF(OUTPUT, gstr)
          resample_field(&infield[k], &inwfield[k], outfield, outwfield,
		prefs.resamp_type);
          }
        }
      }
    }

#ifdef HAVE_MPI
/* Synchronize all nodes */
  MPI_Barrier(MPI_COMM_WORLD);

/* Exit if computing is distributed and it is not the main node */
  if (prefs.nnodes>1 && prefs.node_index)
    goto the_end;
#endif

  if (!prefs.combine_flag)
    goto the_end;

/* Apply flux scaling to input images */
  w1 = w2 = invscale = 0.0;
  for (k=0; k<ntinfield; k++)
    {
    infield[k]->cat->tab->bscale *= infield[k]->fscale;
    infield[k]->backmean *= infield[k]->fscale;
    infield[k]->backsig *= infield[k]->fscale;
    infield[k]->gain /= infield[k]->fscale;
/*--- Update gain information */
     w = 1/(infield[k]->backsig*infield[k]->backsig);
    w1 += w;
    if (infield[k]->gain)
      w2 += w/infield[k]->gain;
    invscale += 1/infield[k]->fscale;
    }

/* Approximation to the final equivalent gain */
  if (w2)
    outfield->gain = (w1/w2)*invscale;
  else
    outfield->gain = 0.0;

/* Add relevant information to output FITS headers */
  writefitsinfo_outfield(outfield, *infield);
  writefitsinfo_outfield(outwfield, inwfield? *inwfield : *infield);

/* Go! */
  coadd_fields(infield, inwfield, ntinfield, outfield, outwfield,
		prefs.coadd_type, BIG);

/* Close files and free memory */
the_end:
  NFPRINTF(OUTPUT, "Closing files...")
  for (k=0; k<ntinfield; k++)
    {
    end_field(infield[k]);
    if (inwfield[k])
      end_field(inwfield[k]);
    }
  free(infield);
  free(inwfield);

  end_field(outfield);
  end_field(outwfield);
  cleanup_files();

  return;
  }


#ifdef HAVE_MPI
/****** control_mpi *******************************************************
PROTO	void control_mpi(int nnodes, int ntasks)
PURPOSE	Control task operations across nodes and provide load balancing.
INPUT	Total number of nodes involved in the processing.
OUTPUT	-.
NOTES	This is your standard Master-Slave control function. If a node does not
	respond at the beginning, it is ignored for the rest of the processing.
AUTHOR	E. Bertin (IAP)
VERSION	08/05/2003
 ***/
void	control_mpi(int nnodes, int ntasks)
  {
    MPI_Request	*request;
    int		*nindex,
		n, t, node, ndone, dummy;

  QMALLOC(request, MPI_Request, nnodes);
  QMALLOC(nindex, int, nnodes);
  for (n=1; n<nnodes; n++)
    MPI_Irecv(&dummy, 1, MPI_INT, n, n, MPI_COMM_WORLD, &request[n]);

  for (t=0; t<ntasks;)
    {
    MPI_Waitsome(nnodes-1, request+1, &ndone, nindex, MPI_STATUSES_IGNORE);
    for (n=0; n<ndone; n++)
      {
      node = nindex[n]+1;
      if (t<ntasks)
        {
        MPI_Send(&t, 1 , MPI_INT, node, node, MPI_COMM_WORLD);
	printf("(%d)  %d --> %d\n", ndone, node, t);
        MPI_Irecv(&dummy, 1, MPI_INT, node, node, MPI_COMM_WORLD,
		&request[node]);
        t++;
	}
      else
        {
        dummy = -1;
        MPI_Send(&dummy, 1 , MPI_INT, node, node, MPI_COMM_WORLD);
        }
      }
    }

/* Wait for remaining active communications to complete */
  dummy = -1;
  for (n=1; n<nnodes;)
    MPI_Send(&dummy, 1 , MPI_INT, n, n, MPI_COMM_WORLD);

  free(request);
  free(nindex);

/* Synchronize all nodes */
  MPI_Barrier(MPI_COMM_WORLD);

  return;
  }
#endif

